/*
 Wireless Gameboy/GBA Controller - Multi-RX (USB, NES, SNES, GC, Wii)
 Version: 1.1
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 3/10/2018
 Last Modified: 10/04/2019
 
 Use any Gameboy / GBA which supports a Gameboy or GBA cartridge such as the Gameboy, MGB, GBC, GBA, GBA SP, 
 DS, DS Lite and GB Micro as a Wireless game controller.
 
 Using V-USB for USB
 WiiMote code provided by the NES2Wii Project & WiiMote Extension Library.
 
 
 Program the ATmega168PB to use a 12MHz crystal
 avrdude -p ATmeg168PB -c usbasp -U lfuse:w:0xEF:m -U hfuse:w:0xDF:m
 
*/

// ATmega168PB Pin map
//
// PB0 Wiimote SCL, GC Data In, NES/SNES Clock		PC0 SNES Enable			
// PB1 nRF CE													PC1 GC Enable
// PB2 nRF CSN													PC2 WiiMote Enable									PD2 USB D+
// PB3 nRF MOSI												PC3 NES/SNES Latch									PD3 USB D-
// PB4 nRF MISO												PC4 Wiimote SDA, GC Data Out, NES/SNES Data	PD4 Wiimote SCL, GC Data In, NES/SNES Clock
// PB5 nRF SCK													PC5 Wiimote SCL, GC Data In, NES/SNES Clock
// PB6 12MHz XTAL
// PB7 12MHz XTAL																											PD7 LED


#define F_CPU 12000000

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <string.h>
#include <avr/sleep.h>
#include <math.h>
#include <util/delay.h>
#include "usbdrv/usbdrv.h"
#include "usbdrv/oddebug.h"
#include "setup.c"
#include "mirf.c"
#include "defines.h"
#include "wiimote.c"

uint8_t modeFound = 0;
uint16_t packetTimeout = 0;
uint8_t kpData = 0;
uint8_t kpDataHigh = 0;
uint16_t volatile kpOutput = 0xFFFF;
uint8_t firstPacket = 1;
uint8_t kpOutputMapYB = 0;
uint8_t kpOutputMapYA = 0;
uint8_t kpOutputMapRtoA = 0;
uint8_t kpOutputMapLtoA = 0;
uint8_t kpOutputUseJoystick = 0;
uint8_t volatile timedOut = 0;

// Gamecube
uint8_t volatile gcInputBuffer[10];
uint8_t kpOutputDpad = 0;
uint8_t kpOutputZButton = 0;
uint8_t kpOutputLRDownAndClick = 0;

uint8_t kpLHeld = 0;
uint8_t kpLHeldCounter = 0;
uint8_t kpRHeld = 0;
uint8_t kpRHeldCounter = 0;

uint8_t gcReply1[3] = {0x09, 0x00, 0x20};
uint8_t gcReply2[10] = {0x00, 0x80, 0x7C, 0x78, 0x80, 0x79, 0x1A, 0x1B, 0x02, 0x02};
uint8_t gcControllerData[8] = {0x00, 0x80, 0x7D, 0x78, 0x80, 0x7A, 0x1A, 0x1B};

// WiiMote Extension button presses
volatile int8_t jx = 0, jy = 0, rx = 0, ry = 0;
volatile uint8_t tl = 0, tr = 0;
volatile uint8_t left = 0, right = 0, up = 0, down = 0, a = 0, b = 0, x = 0, y = 0, 
						select = 0, start = 0, home = 0, l = 0, r = 0, zl = 0, zr = 0;

void wiimote_query(void) {
	unsigned char but_dat[8]; // struct containing button data
	if (twi_reg[0xFE] == 1) // data format
	{
		but_dat[0] = 0b00000000; // RX<4:3>	LX<5:0>
		but_dat[1] = 0b00000000; // RX<2:1>	LY<5:0>
		but_dat[2] = 0b00000000; // RX<0>	LT<4:3>	RY<4:0>
		but_dat[3] = 0b00000000; // LT<2:0>	RT<4:0>
		but_dat[4] = 0b11111111; // BDR	BDD	BLT	B-	BH	B+	BRT	 1
		but_dat[5] = 0b11111111; // BZL	BB	BY	BA	BX	BZR	BDL	BDU
		but_dat[6] = 0;
		but_dat[7] = 0;
		
		but_dat[0] |= (jx / 4 + 0x20) & 0x3F;
		but_dat[1] |= (jy / -4 + 0x20) & 0x3F;
		but_dat[0] |= ((rx / 8 + 0x10) & 0x18) << 3;
		but_dat[1] |= ((rx / 8 + 0x10) & 0x06) << 5;
		but_dat[2] |= ((rx / 8 + 0x10) & 0x01) << 7;
		but_dat[2] |= (ry / -8 + 0x10) & 0x1F;
		but_dat[2] |= ((tl / 8) & 0x18) << 2;
		but_dat[3] |= ((tl / 8) & 0x07) << 5;
		but_dat[3] |= (tr / 8) & 0x1F;
		
		if (right) but_dat[4] &= ~(1<<7);
		if (down) but_dat[4] &= ~(1<<6);
		if (l) but_dat[4] &= ~(1<<5);
		if (select) but_dat[4] &= ~(1<<4);
		if (home) but_dat[4] &= ~(1<<3);
		if (start) but_dat[4] &= ~(1<<2);
		if (r) but_dat[4] &= ~(1<<1);
		
		if (zl) but_dat[5] &= ~(1<<7);
		if (b) but_dat[5] &= ~(1<<6);
		if (y) but_dat[5] &= ~(1<<5);
		if (a) but_dat[5] &= ~(1<<4);
		if (x) but_dat[5] &= ~(1<<3);
		if (zr) but_dat[5] &= ~(1<<2);
		if (left) but_dat[5] &= ~(1<<1);
		if (up) but_dat[5] &= ~(1<<0);
		
		but_dat[6] = 0;
		but_dat[7] = 0;
		
		wm_newaction(but_dat);
	} 
	else if (twi_reg[0xFE] == 3) // data format
	{
		but_dat[0] = jx + 0x80;
		but_dat[1] = rx + 0x80;
		but_dat[2] = 0x7fl - jy;
		but_dat[3] = 0x7fl - ry;
		but_dat[4] = tl;
		but_dat[5] = tr;
		
		but_dat[6] = 0b11111111; // BDR	BDD	BLT	B-	BH	B+	BRT	 1
		but_dat[7] = 0b11111111; // BZL	BB	BY	BA	BX	BZR	BDL	BDU
		
		if (right) but_dat[6] &= ~(1<<7);
		if (down) but_dat[6] &= ~(1<<6);
		if (l) but_dat[6] &= ~(1<<5);
		if (select) but_dat[6] &= ~(1<<4);
		if (home) but_dat[6] &= ~(1<<3);
		if (start) but_dat[6] &= ~(1<<2);
		if (r) but_dat[6] &= ~(1<<1);
		
		if (zl) but_dat[7] &= ~(1<<7);
		if (b) but_dat[7] &= ~(1<<6);
		if (y) but_dat[7] &= ~(1<<5);
		if (a) but_dat[7] &= ~(1<<4);
		if (x) but_dat[7] &= ~(1<<3);
		if (zr) but_dat[7] &= ~(1<<2);
		if (left) but_dat[7] &= ~(1<<1);
		if (up) but_dat[7] &= ~(1<<0);
		
		wm_newaction(but_dat);
	} 
}

// Gamecube code
// While high, count bit time. Under 1.5uS low, higher than 2us high
uint8_t read_bit(void) {
	uint8_t bitTime = 0;
	
	while (!(PINC & (1<<GC_DATA_IN))); // Wait until high
	//PORTD |= (1<<LED);
	asm volatile("nop");
	
	while (PINC & (1<<GC_DATA_IN)) { // While high, count
		bitTime++;
	}
	asm volatile("nop");
	//PORTD &= ~(1<<LED);
	
	if (bitTime >= 6) { // High
		return 1;
	}
	else {
		return 0; // Low
	}
}

// Read data to the array until the count. Use the bypassHighLow if running this after interrupt
void read_bytes(uint8_t count) {
	while (!(PINC & (1<<GC_DATA_IN))); // Wait until high
	while (PINC & (1<<GC_DATA_IN)); // Wait until low
	
	for (uint8_t byte = 0; byte < count; byte++) {
		uint8_t dataIn = 0;
		
		for (uint8_t x = 0x80; x != 0; x >>= 1) { // Loop 8 times (fast)
			if (read_bit()) {
				dataIn |= x;
			}
		}
		gcInputBuffer[byte] = dataIn;
	}
	
	while (!(PINC & (1<<GC_DATA_IN))); // Wait until high
	_delay_us(1);
}

// Write data from the array until the count
void write_bytes(uint8_t* inputArray, uint8_t count) {
	for (uint8_t byte = 0; byte < count; byte++) {
		for (uint8_t x = 0x80; x != 0; x >>= 1) { // Loop 8 times
			if (inputArray[byte] & x) {
				PORTC |= (1<<GC_DATA_OUT);
				asm volatile("nop");
				asm volatile("nop");
				asm volatile("nop");
				asm volatile("nop");
				asm volatile("nop");
				PORTC &= ~(1<<GC_DATA_OUT);
				_delay_us(3);
				PORTC |= (1<<GC_DATA_OUT);
			}
			else {
				PORTC |= (1<<GC_DATA_OUT);
				_delay_us(3);
				PORTC &= ~(1<<GC_DATA_OUT);
				_delay_us(1);
				PORTC |= (1<<GC_DATA_OUT);
			}
		}
	}
	PORTC |= (1<<GC_DATA_OUT);
	_delay_us(2);
	PORTC &= ~(1<<GC_DATA_OUT);
}


int main (void) {
	setup();
	
	// Init the adapter
	if (adapterConnected == USB_CONNECTED) {
		usbInit(); // Start USB
	}
	else if (adapterConnected == GC_CONNECTED) {
		gcInputBuffer[0] = 0;
	}
	else if (adapterConnected == WII_CONNECTED) {
		wm_init((void*)classic_controller_id, (void*)cal_data, wiimote_query); // Init as a controller
	}
	
	// Check for new nRF settings for 0.5 seconds
	check_for_new_settings(); 
	
	// Check for GB/GBA mode
	check_for_gb_or_gba_mode();
	
	// Turn on time out counter for ~5 seconds
	TCNT1 = 0;
	TIMSK1 = (1<<TOV1);
	TCCR1B = (1<<CS12) | (1<<CS10);
	
	// Wait for latch low for NES/SNES
	if (adapterConnected == NES_CONNECTED || adapterConnected == SNES_CONNECTED) {
		while (!(PINC & (1<<NES_SNES_LATCH))); // Wait until high
		_delay_us(500);
		
		// Turn on pin interrupt on latch pin
		PCMSK1 |= (1<<PCINT11);
		PCICR |= (1<<PCIE1);
	}
	
	
	while(1) {
		if (timedOut == 1) { // If timed out, check for the transmitter again
			// Reset to defaults
			kpOutput = 0xFFFF;
			reportBuffer.data = 0x00;
			left = 0, right = 0, up = 0, down = 0, a = 0, b = 0, x = 0, y = 0, 
			select = 0, start = 0, home = 0, l = 0, r = 0, zl = 0, zr = 0;
			gcControllerData[0] = 0;
			gcControllerData[1] = 0x80;
			gcControllerData[2] = 0x7D;
			gcControllerData[3] = 0x78;
			gcControllerData[6] = 0x1A;
			gcControllerData[7] = 0x1B;
			kpLHeld = 0;
			kpRHeld = 0;
			
			check_for_gb_or_gba_mode();
			timedOut = 0;
			TCNT1 = 0;
			TCCR1B = (1<<CS12) | (1<<CS10);
		}
		
		if (mirf_status() & (1<<RX_DR)) { // Packet received
			PORTD |= (1<<LED);
			mirf_CE_lo; // Stop listening
			
			mirf_CSN_lo; // Pull down chip select
			spi_transfer(R_RX_PAYLOAD); // Send cmd to read rx payload
			
			if (modeFound == GBA_MODE) { // Load 2 bytes if in GBA mode
				kpDataHigh = spi_transfer(0);
				kpData = spi_transfer(0);
			}
			else {
				kpData = spi_transfer(0); // Read data
			}
			
			mirf_CSN_hi; // Pull up chip select
			mirf_config_register(STATUS,(1<<RX_DR)); // Reset status register
			
			mirf_flush_rx_tx(); // Flush TX/RX
			mirf_CE_hi; // Start listening
			PORTD &= ~(1<<LED);
			
			
			if (adapterConnected == NES_CONNECTED) {
				// Check for key presses
				uint16_t tempKpOutput = 0xFFFF;
				if (kpData & 0x01) { // A
					tempKpOutput &= ~(1<<7);
				}
				if (kpData & 0x02) { // B
					tempKpOutput &= ~(1<<6);
				}
				if (kpData & 0x04) { // Right
					tempKpOutput &= ~(1<<0);
				}
				if (kpData & 0x08) { // Left
					tempKpOutput &= ~(1<<1);
				}
				if (kpData & 0x10) { // Up
					tempKpOutput &= ~(1<<3);
				}
				if (kpData & 0x20) { // Down
					tempKpOutput &= ~(1<<2);
				}
				if (kpData & 0x40) { // Select
					tempKpOutput &= ~(1<<5);
				}
				if (kpData & 0x80) { // Start
					tempKpOutput &= ~(1<<4);
				}
				
				// Assign new key presses
				kpOutput = tempKpOutput;
			}
			else if (adapterConnected == SNES_CONNECTED) {
				// The first packet received triggers special modes
				if (firstPacket == 1) {
					firstPacket = 0;
					
					// A held down, map the Gameboy A button to the SNES Y button and Gameboy B button to the SNES A button
					if (kpData & 0x01) { // A
						kpOutputMapYB = 1;
					}
					// L held down, map GBA L button to the SNES A button
					if (kpDataHigh & 0x01) {
						kpOutputMapLtoA = 1;
					}
					// R held down, map GBA R button to the SNES A button
					if (kpDataHigh & 0x02) {
						kpOutputMapRtoA = 1;
					}
				}
				
				// Check for key presses
				uint16_t tempKpOutput = 0xFFFF;
				if (kpData & 0x01) { // A
					if (kpOutputMapYB == 1) {
						tempKpOutput &= ~(1<<15);
					}
					else {
						tempKpOutput &= ~(1<<7);
					}
				}
				if (kpData & 0x02) { // B
					if (kpOutputMapYB == 1) {
						tempKpOutput &= ~(1<<14);
					}
					else {
						tempKpOutput &= ~(1<<15);
					}
				}
				if (kpData & 0x04) { // Right
					tempKpOutput &= ~(1<<8);
				}
				if (kpData & 0x08) { // Left
					tempKpOutput &= ~(1<<9);
				}
				if (kpData & 0x10) { // Up
					tempKpOutput &= ~(1<<11);
				}
				if (kpData & 0x20) { // Down
					tempKpOutput &= ~(1<<10);
				}
				if (kpData & 0x40) { // Select
					tempKpOutput &= ~(1<<13);
				}
				if (kpData & 0x80) { // Start
					tempKpOutput &= ~(1<<12);
				}
				if (kpDataHigh & 0x01) { // L (for GBA)
					if (kpOutputMapLtoA == 1) {
						tempKpOutput &= ~(1<<7);
					}
					else {
						tempKpOutput &= ~(1<<5);
					}
				}
				if (kpDataHigh & 0x02) { // R (for GBA)
					if (kpOutputMapRtoA == 1) {
						tempKpOutput &= ~(1<<7);
					}
					else {
						tempKpOutput &= ~(1<<4);
					}
				}
				
				// Assign new key presses
				kpOutput = tempKpOutput;
			}
			else if (adapterConnected == GC_CONNECTED) {
				// The first packet received triggers special modes
				if (firstPacket == 1) {
					firstPacket = 0;
					
					// Down held down, output the key presses to the D-Pad
					if (kpData & 0x20) {
						kpOutputDpad = 1;
					}
					
					// R held down, simulate the L/R button presses like in the GC controller, if they are held for more than 200ms
					// then also trigger the click switch too. Useful for Gamecube games that might use this.
					if (kpDataHigh & 0x02) { 
						kpOutputLRDownAndClick = 1;
					}
					
					// Select held down, output the select key to the Z button
					if (kpData & 0x40) {
						kpOutputZButton = 1;
					}
					
					// A held down, B button now outputs Y and A button outputs B
					if (kpData & 0x01) { 
						kpOutputMapYB = 1;
					}
					
					// L held down, map GBA L button to the A button
					if (kpDataHigh & 0x01) {
						kpOutputMapLtoA = 1;
					}
				}
				
				// Default values
				uint8_t tempGcControllerData[8] = {0x00, 0x80, 0x7D, 0x78, 0x80, 0x7A, 0x1A, 0x1B};
				kpLHeld = 0;
				kpRHeld = 0;
				
				// Check for key presses
				if (kpData & 0x01) { // A
					if (kpOutputMapYB == 1) { // Map B button to A
						tempGcControllerData[0] |= 0x02;
					}
					else {
						tempGcControllerData[0] |= 0x01;
					}
				}
				if (kpData & 0x02) { // B
					if (kpOutputMapYB == 1) { // Map Y button to B
						tempGcControllerData[0] |= 0x08;
					}
					else {
						tempGcControllerData[0] |= 0x02;
					}
				}
				if (kpData & 0x04) { // Right
					if (kpOutputDpad == 1) {
						tempGcControllerData[1] |= 0x02; // D-Pad
					}
					else {
						tempGcControllerData[2] = 0xFF;
					}
				}
				if (kpData & 0x08) { // Left
					if (kpOutputDpad == 1) {
						tempGcControllerData[1] |= 0x01; // D-Pad
					}
					else {
						tempGcControllerData[2] = 0x00;
					}
				}
				if (kpData & 0x10) { // Up
					if (kpOutputDpad == 1) {
						tempGcControllerData[1] |= 0x08; // D-Pad
					}
					else {
						tempGcControllerData[3] = 0xFF;
					}
				}
				if (kpData & 0x20) { // Down
					if (kpOutputDpad == 1) {
						tempGcControllerData[1] |= 0x04; // D-Pad
					}
					else {
						tempGcControllerData[3] = 0x00;
					}
				}
				if (kpData & 0x40) { // Select
					if (kpOutputZButton == 0) {
						tempGcControllerData[0] |= 0x08; // Y
					}
					else {
						tempGcControllerData[1] |= 0x10; // Z
					}
				}
				if (kpData & 0x80) { // Start
					tempGcControllerData[0] |= 0x10;
				}
				if (kpDataHigh & 0x01) { // L (for GBA)
					if (kpOutputLRDownAndClick == 1) {
						kpLHeld = 1;
						kpLHeldCounter++;
						tempGcControllerData[6] = 0x78;
						
						if (kpLHeldCounter >= 250) {
							tempGcControllerData[1] |= 0x40;
							kpLHeldCounter = 250;
						}
					}
					else if (kpOutputMapLtoA == 1) {
						tempGcControllerData[0] |= 0x01;
					}
					else {
						tempGcControllerData[1] |= 0x40;
					}
				}
				if (kpDataHigh & 0x02) { // R (for GBA)
					if (kpOutputLRDownAndClick == 1) {
						kpRHeld = 1;
						kpRHeldCounter++;
						
						tempGcControllerData[7] = 0x78;
						if (kpRHeldCounter >= 250) {
							tempGcControllerData[1] |= 0x20;
							kpRHeldCounter = 250;
						}
					}
					else {
						tempGcControllerData[1] |= 0x20;
					}
				}
				
				// Reset L/R counter if not held down
				if (kpOutputLRDownAndClick == 1) {
					if (kpLHeld == 0) {
						kpLHeldCounter = 0;
					}
					if (kpRHeld == 0) {
						kpRHeldCounter = 0;
					}
				}
				
				// Assign new key presses
				for (uint8_t x = 0; x < 8; x++) {
					gcControllerData[x] = tempGcControllerData[x];
				}
			}
			else if (adapterConnected == WII_CONNECTED) {
				// The first packet received triggers special modes
				if (firstPacket == 1) {
					firstPacket = 0;
					
					// A held down, map the Gameboy A button to the SNES Y button and Gameboy B button to the SNES A button
					if (kpData & 0x01) {
						kpOutputMapYB = 1;
					}
					// B held down, map the Gameboy A button to the SNES Y button and Gameboy B button to the SNES A button
					if (kpData & 0x02) {
						kpOutputMapYA = 1;
					}
					// L held down, map GBA L button to the SNES A button
					if (kpDataHigh & 0x01) {
						kpOutputMapLtoA = 1;
					}
					// R held down, map GBA R button to the SNES A button
					if (kpDataHigh & 0x02) {
						kpOutputMapRtoA = 1;
					}
					// Down held down, use the joystick instead of the DPad
					if (kpData & 0x01) { // A
						kpOutputUseJoystick = 1;
					}
				}
				
				// Turn off interrupts before resetting key press data and assigning values
				cli();
				
				// Reset key presses
				left = 0, right = 0, up = 0, down = 0, a = 0, b = 0, x = 0, y = 0, 
				select = 0, start = 0, home = 0, l = 0, r = 0, zl = 0, zr = 0;
				
				// Check for key presses
				if (kpData & 0x01) { // A
					a = 1;
				}
				if (kpData & 0x02) { // B
					b = 1;
				}
				if (kpData & 0x04) { // Right
					right = 1;
				}
				if (kpData & 0x08) { // Left
					left = 1;
				}
				if (kpData & 0x10) { // Up
					up = 1;
				}
				if (kpData & 0x20) { // Down
					down = 1;
				}
				if (kpData & 0x40) { // Select
					select = 1;
				}
				if (kpData & 0x80) { // Start
					start = 1;
				}
				if (kpDataHigh & 0x01) { // L (for GBA)
					l = 1;
				}
				if (kpDataHigh & 0x02) { // R (for GBA)
					r = 1;
				}
				
				// Turn interrupts back on
				sei();
			}
		}
		
		
		// USB Poll
		if (adapterConnected == USB_CONNECTED) {
			usbPoll();
			
			if (usbInterruptIsReady()) {
				reportBuffer.data = 0x00; // Center pad
				
				// Check for key presses
				if (kpData & 0x01) { // A
					reportBuffer.A = 1;
				}
				if (kpData & 0x02) { // B
					reportBuffer.B = 1;
				}
				if (kpData & 0x04) { // Right
					reportBuffer.X_JOY = 127;
				}
				if (kpData & 0x08) { // Left
					reportBuffer.X_JOY = -127;
				}
				if (kpData & 0x10) { // Down
					reportBuffer.Y_JOY = -127;
				}
				if (kpData & 0x20) { // Up
					reportBuffer.Y_JOY = 127;
				}
				if (kpData & 0x40) { // Select
					reportBuffer.SELECT = 1;
				}
				if (kpData & 0x80) { // Start
					reportBuffer.START = 1;
				}
				if (kpDataHigh & 0x01) { // L (for GBA)
					reportBuffer.L = 1;
				}
				if (kpDataHigh & 0x02) { // R (for GBA)
					reportBuffer.R = 1;
				}
				
				// Called after every poll of the interrupt endpoint
				usbSetInterrupt(&reportBuffer, sizeof(reportBuffer));
			}
		}
		else if (adapterConnected == GC_CONNECTED) {
			// At first boot and if we ever receive 0 as the first byte (e.g when GC reboots)
			if (gcInputBuffer[0] == 0) {
				PCICR &= ~(1<<PCIE2); // Turn off interrupts
				//_delay_ms(2000);
				
				// Wait for initial request from Gamecube (0000 0000)
				// Transmit Gamecube controller reply 1
				read_bytes(1);
				while (gcInputBuffer[0] != 0) {
					_delay_ms(10);
					read_bytes(1);
				}
				write_bytes(gcReply1, 3);
				
				
				// Gamecube responds (0100 0001)
				// Transmit Gamecube controller reply 2
				read_bytes(1);
				write_bytes(gcReply2, 10);
				
				// Turn on pin interrupt
				_delay_us(5); // Wait a little
				PCMSK2 |= (1<<PCINT20);
				PCICR |= (1<<PCIE2);
			}
		}
		
		//PORTD |= (1<<LED);
		_delay_us(250);
	}
	
	return 0;
}


// NES/SNES - Latch pin was toggled
ISR(PCINT1_vect) {
	if (!(PINC & (1<<NES_SNES_LATCH))) { // Latch low
		PORTD &= ~(1<<LED);
		uint8_t latchReset = 0;
		
		// 8/16 bits of data to write depending on what keys were pressed
		for (uint16_t x = loopTimes; x != 0; x >>= 1) { // Loop 8/16 times
			if (kpOutput & x) {
				PORTC |= (1<<NES_SNES_DATA);
			}
			else {
				PORTC &= ~(1<<NES_SNES_DATA);
			}
			
			while (PINC & (1<<NES_SNES_CLK)) { // Wait for clock low
				if (PINC & (1<<NES_SNES_LATCH)) { // Latch is high, break out of everything
					latchReset = 1;
					break;
				}
			}
			if (latchReset == 1) {
				break;
			}
			while (!(PINC & (1<<NES_SNES_CLK))); // Wait for clock high
		}
		PORTC &= ~(1<<NES_SNES_DATA); // Set data low (required for SNES)
	}
	else { // Latch high, set first bit
		if (kpOutput & loopTimes) {
			PORTC |= (1<<NES_SNES_DATA);
		}
		else {
			PORTC &= ~(1<<NES_SNES_DATA);
		}
	}
	_delay_us(5);
}

// GC - Data input was toggled
ISR(PCINT2_vect) {
	if (!(PINC & (1<<GC_DATA_IN))) {
		PORTD &= ~(1<<LED);
		
		for (uint8_t byte = 0; byte < 3; byte++) {
			uint8_t dataIn = 0;
			for (uint8_t x = 0x80; x != 0; x >>= 1) { // Loop 8 times (fast)
				uint8_t bitTime = 0;
				
				while (!(PINC & (1<<GC_DATA_IN))); // Wait until high
				//PORTD |= (1<<LED);
				asm volatile("nop");
				
				while (PINC & (1<<GC_DATA_IN)) { // While high, count
					bitTime++;
				}
				//PORTD &= ~(1<<LED);
				asm volatile("nop");
				
				if (bitTime >= 6) { // High
					dataIn |= x;
				}
			}
			gcInputBuffer[byte] = dataIn;
		}
		
		while (!(PINC & (1<<GC_DATA_IN))); // Wait until high
		_delay_us(1);
		
		// Write the controller data
		for (uint8_t byte = 0; byte < 8; byte++) {
			for (uint8_t x = 0x80; x != 0; x >>= 1) { // Loop 8 times (fast)
				if (gcControllerData[byte] & x) {
					PORTC |= (1<<GC_DATA_OUT);
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					PORTC &= ~(1<<GC_DATA_OUT);
					_delay_us(3);
					PORTC |= (1<<GC_DATA_OUT);
				}
				else {
					PORTC |= (1<<GC_DATA_OUT);
					_delay_us(3);
					PORTC &= ~(1<<GC_DATA_OUT);
					_delay_us(1);
					PORTC |= (1<<GC_DATA_OUT);
				}
			}
		}
		PORTC |= (1<<GC_DATA_OUT);
		_delay_us(2);
		PORTC &= ~(1<<GC_DATA_OUT);
		
		_delay_us(5);
	}
}


// Overflows after ~5 seconds of no packets received
ISR(TIMER1_OVF_vect) {
	timedOut = 1;
	TCNT1 = 0;
	TCCR1B = 0;
}
